'''MOCCASIN: Model ODE Converter for Creating Automated SBML INteroperability

Synopsis
--------

MOCCASIN is designed to convert certain basic forms of ODE simulation models
written in MATLAB or Octave and translate them into SBML format.


Background and introduction
---------------------------

Computational modeling has become a crucial aspect of biological research,
and SBML (the Systems Biology Markup Language) has become the de facto
standard open format for exchanging models between software tools in systems
biology. MATLAB and Octave are popular numerical computing environments used
by modelers in biology, but while toolboxes for using SBML exist, many
researchers either have legacy models or do not learn about the toolboxes
before starting their work and then find it discouragingly difficult to
export their MATLAB/Octave models to SBML.

The goal of this project is to develop software that uses a combination of
heuristics and user assistance to help researchers export models written as
ordinary MATLAB and Octave scripts. MOCCASIN ("Model ODE Converter for
Creating Automated SBML INteroperability") helps researchers take ODE
(ordinary differential equation) models written in MATLAB and Octave and
export them as SBML files. Although its scope is limited to MATLAB written
with certain assumptions, and general conversion of MATLAB models is
impossible, MOCCASIN nevertheless can translate some common forms of models
into SBML.

MOCCASIN is written in Python and does not require MATLAB to run. It requires
libSBML and a number of common Python 3 libraries to run.


How it works
------------

MOCCASIN uses an algorithm developed by Fages, Gay and Soliman described in
their 2015 paper titled "Inferring reaction systems from ordinary
differential equations" (Theoretical Computer Science, vol 599, pp. 64-78).
To parse MATLAB and produce input to the reaction-inference algorithm,
MOCCASIN uses a custom MATLAB parser written using PyParsing and a variety of
post-processing operations to interpret the MATLAB contents.

Currently, MOCCASIN is limited to MATLAB inputs in which a model is contained
in a single file. The file must set up a system of differential equations as
a function defined in the file, and make a call to one of the MATLAB odeNN
family of solvers (e.g., ode45, ode15s, etc.). The following is a simple but
complete example:

  % Various parameter settings.  The specifics here are unimportant; this
  % is just an example of a real input file.
  %
  tspan  = [0 300];
  xinit  = [0; 0];
  a      = 0.01 * 60;
  b      = 0.0058 * 60;
  c      = 0.006 * 60;
  d      = 0.000192 * 60;

  % A call to a MATLAB ODE solver
  %
  [t, x] = ode45(@f, tspan, xinit);

  % A function that defines the ODEs of the model.
  %
  function dx = f(t, x)
    dx = [a - b * x(1); c * x(1) - d * x(2)];
  end

MOCCASIN assumes that the second parameter in the ODE function definition
determines the variables that should identify the SBML species; thus, the
output generated by MOCCASIN will have SBML species named x_1 and x_2 by
default. (The use of suffixes is necessary because plain SBML does not
support arrays or vectors.) The output will also not capture any information
about the particular ODE solver or the start/stop/configuration parameters
used in the file, because that kind of information is not meant to be stored
in SBML files anyway. (A future version of MOCCASIN will hopefully translate
the additional run information into SED-ML format.)


Using MOCCASIN
--------------

You can use MOCCASIN either via the command line or via a built-in GUI
interface.  Both interface area accessed using the same command, with the
default being to start the GUI interface.  On Linux and macOS systems, the
installation _should_ place a new program on your shell's search path, so
that you can start MOCCASIN with a simple shell command:

  moccasin

If that fails because the shell cannot find the command, you should be able
to run it using the alternative approach:

  python3 -m moccasin

To use the command-line interface, supply one or more MATLAB files on the
command line; MOCCASIN will read and convert the file(s):

  moccasin path/to/a/matlab/file.m

It also accepts additional command-line arguments.  To get information about
the other options, use the -h argument:

  moccasin -h


Authors
-------

Michael Hucka <mhucka@caltech.edu>
Sarah M. Keating <skeating@caltech.edu>
Harold Gomez <harold.gomez@bsse.ethz.ch>


Copyright
---------

This software is part of MOCCASIN, the Model ODE Converter for Creating
Automated SBML INteroperability. Visit https://github.com/sbmlteam/moccasin/.

Copyright (C) 2014-2018 jointly by the following organizations:
    1. California Institute of Technology, Pasadena, CA, USA
    2. Icahn School of Medicine at Mount Sinai, New York, NY, USA
    3. Boston University, Boston, MA, USA

This is free software; you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free
Software Foundation.  A copy of the license agreement is provided in the
file named "COPYING.txt" included with this software distribution and also
available online at https://github.com/sbmlteam/moccasin/.
'''

from .__version__ import __version__, __title__, __url__, __description__
from .__version__ import __author__, __email__, __citation__
from .__version__ import __license__, __license_url__, __help_url__

from .matlab_parser import MatlabParser
from .interfaces import moccasin_CLI, moccasin_GUI
from .converter import create_raterule_model, process_biocham_output
from .converter import sanity_check_matlab
from .errors import *
